structure IL3Static :> IL3STATIC =
  struct
    open IL3
    structure Ctx :> CONTEXT
                      where type con = con
                        and type kind = kind
                  =
         ContextFn(struct
                    type con = con
                    type kind = kind
                   end)

    type context = Ctx.context

    exception Equiv
    exception Check of string
    exception Error of string

    fun check b f = if b then () else f ()

    fun nth (l, n) = SOME (List.nth (l,n)) handle Subscript => NONE

    fun option_map f opt =
        case opt of
            NONE => NONE
          | SOME x => SOME (f x)

    (*** static semantics ***)

    fun equiv_kind knd1 knd2 =
        case (knd1, knd2) of
            (KndType, KndType) => true
          | (KndArrow (k1dom, k1cod), KndArrow (k2dom, k2cod)) =>
                equiv_kind k1dom k2dom
                andalso
                equiv_kind k1cod k2cod
          | (KndUnit, KndUnit) => true
          | (KndPair (k1fst, k1snd), KndPair (k2fst, k2snd)) =>
                equiv_kind k1fst k2fst
                andalso
                equiv_kind k1snd k2snd
          | _ => false

    fun whnf con =
        let fun reduce con =
                case con of
                  (* beta reductions *)
                    ConApp (ConAbs (cx, k, c1), c2) =>
                        SOME (subst_con_con c2 cx c1)
                  | ConProj1 (ConPair (c1, _)) => SOME c1
                  | ConProj2 (ConPair (_, c2)) => SOME c2
                  (* congruences *)
                  | ConApp (c1, c2) => option_map (fn c1' => ConApp (c1', c2))
                                                  (reduce c1)
                  | ConProj1 c => option_map (fn c' => ConProj1 c') (reduce c)
                  | ConProj2 c => option_map (fn c' => ConProj2 c') (reduce c)
                  | _ => NONE
        in
            case reduce con of
                SOME con' => whnf con'
              | NONE => con
        end

    (* arguably could be done functionally now, without exceptions, since
       there are no more evars to side effectually update *)
    and equiv_con ctx con1 con2 knd =
        case knd of
            KndType =>
                let val k = equiv_con_path ctx (whnf con1) (whnf con2)
                    (* is this kind check necessary?  maybe for the sake of
                       belt and suspenders... *)
                    val () = if equiv_kind k KndType then
                                ()
                             else
                                raise Error
                             ("mismatch between equiv_con input (KndType) "
                             ^ " and equiv_con_path output")
                in
                    ()
                end
          | KndUnit => () (* con1 ≡ con2 if con1 : 1 and con2 : 1 *)
          | KndPair (k1, k2) =>
                let val () = equiv_con ctx (ConProj1 con1) (ConProj1 con2) k1
                    val () = equiv_con ctx (ConProj2 con1) (ConProj2 con2) k2
                in
                    ()
                end
          | KndArrow (k, k') =>
                let val cx = Variable.newNamed "a"
                in
                    equiv_con (Ctx.extend ctx (Ctx.DecConKind (cx, k)))
                              (ConApp (con1, ConVar cx))
                              (ConApp (con2, ConVar cx))
                              k'
                end

    and equiv_con_path ctx con1 con2 =
        case (con1, con2) of
            (ConVar cx1, ConVar cx2) =>
                if Variable.eq (cx1, cx2) then
                    case Ctx.lookup_convar ctx cx1 of
                        NONE => raise Error ("comparing unbound variable, "
                                            ^ Variable.toString cx1)
                      | SOME k => k
                else
                    raise Equiv
          | (ConApp (cp1, c1), ConApp (cp2, c2)) =>
                let val k = equiv_con_path ctx cp1 cp2
                in
                    case k of
                        KndArrow (kdom, kcod) =>
                            let val () = equiv_con ctx c1 c2 kdom
                            in
                                kcod
                            end
                      | _ => raise Error
                             "equiv_con_path: application of non-function"
                end
          | (ConProj1 cp1, ConProj1 cp2) =>
                let val k = equiv_con_path ctx cp1 cp2
                in
                    case k of
                        KndPair (kfst, ksnd) => kfst
                      | _ => raise Error
                             "equiv_con_path: first projection from a non-pair"
                end
          | (ConProj2 cp1, ConProj2 cp2) =>
                let val k = equiv_con_path ctx cp1 cp2
                in
                    case k of
                        KndPair (kfst, ksnd) => ksnd
                      | _ => raise Error
                             "equiv_con_path: second projection from a non-pair"
                end
          | (ConMu (cx1, k1, c1), ConMu (cx2, k2, c2)) =>
                let val () = check (equiv_kind k1 k2) (fn () => raise Equiv)
                    val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2)
                                                     replace_con_con
                    val () = equiv_con (Ctx.extend ctx (Ctx.DecConKind (cx,k1)))
                                       c1 c2 k1
                in
                    k1
                end

          | (TyLit lt1, TyLit lt2) =>
                let val () = check (Literal.lit_ty_eq (lt1, lt2))
                                   (fn () => raise Equiv)
                in
                    KndType
                end
          | (TyArrow (c1dom, c1cod), TyArrow (c2dom, c2cod)) =>
                let val () = equiv_con ctx c1dom c2dom KndType
                    val () = equiv_con ctx c1cod c2cod KndType
                in
                    KndType
                end
          | (TyRef c1, TyRef c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                in
                    KndType
                end
          | (TyTagged, TyTagged) => KndType
          | (TyTag c1, TyTag c2) =>
                let val () = equiv_con ctx c1 c2 KndType
                in
                    KndType
                end
          | (TySum cs1, TySum cs2) =>
                let val () = check (List.length cs1 = List.length cs2)
                                   (fn () => raise Equiv)
                    val () = ListPair.app
                                (fn (c1, c2) => equiv_con ctx c1 c2 KndType)
                                (cs1, cs2)
                in
                    KndType
                end
          | (TyProd cs1, TyProd cs2) =>
                let val () = check (List.length cs1 = List.length cs2)
                                   (fn () => raise Equiv)
                    val () = ListPair.app
                                (fn (c1, c2) => equiv_con ctx c1 c2 KndType)
                                (cs1, cs2)
                in
                    KndType
                end
          | (TyForall (cx1, k1, c1), TyForall (cx2, k2, c2)) =>
                let val () = check (equiv_kind k1 k2) (fn () => raise Equiv)
                    val (cx, (c1, c2)) = alpha_vary2 (cx1, c1) (cx2, c2)
                                                     replace_con_con
                    val () = equiv_con (Ctx.extend ctx (Ctx.DecConKind (cx,k1)))
                                       c1 c2 KndType
                 in
                     KndType
                 end

          | (ConAbs _, ConAbs _) => raise Error "equiv_con_path on non-path"
          | (ConPair _, ConPair _) => raise Error "equiv_con_path on non-path"
          | _ => raise Equiv


    and check_equiv_con ctx con1 con2 knd s =
        equiv_con ctx con1 con2 knd
        handle Equiv => raise Check s


    fun synth_con_kind ctx con =
        case con of
            ConVar cx => 
                (case Ctx.lookup_convar ctx cx of
                    SOME k => k
                  | NONE => raise Check ("synth_con_kind: "
                            ^ "unbound convar, " ^ Variable.toString cx))
          | ConAbs (cx, k, c) =>
                let val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val k' = synth_con_kind ctx c
                in
                    KndArrow (k, k')
                end
          | ConApp (c1, c2) =>
                let val k1 = synth_con_kind ctx c1
                in
                    case k1 of
                        KndArrow (k, k') =>
                            let val () = check_con_kind ctx c2 k
                                         ("synth_con_kind: "
                                         ^ "argument in ConApp has wrong kind")
                            in
                                k'
                            end
                      | _ => raise Check
                             "synth_con_kind: ConApp of non KndArrow"
                end
          | ConUnit => KndUnit
          | ConPair (c1, c2) =>
                KndPair (synth_con_kind ctx c1, synth_con_kind ctx c2)
          | ConProj1 c =>
                let val k = synth_con_kind ctx c
                in
                    case k of
                        KndPair (k1, _) => k1
                      | _ => raise Check
                                   "synth_con_kind: ConProj1 from a non-pair"
                end
          | ConProj2 c =>
                let val k = synth_con_kind ctx c
                in
                    case k of
                        KndPair (_, k2) => k2
                      | _ => raise Check
                                   "synth_con_kind: ConProj2 from a non-pair"
                end
          | ConMu (cx, k, c) =>
                let fun is_Tn k =
                        case k of
                            KndPair (KndType, KndUnit) => true
                          | KndPair (KndType, k) => is_Tn k
                          | _ => false
                    val () = check (is_Tn k) (fn () => raise Check
                             "synth_con_kind: ConMu over non-T^n kind")
                    val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val () = check_con_kind ctx c k
                             "synth_con_kind: ConMu body has wrong kind"
                in
                    k
                end
          | TyLit _ => KndType
          | TyArrow (c1, c2) =>
                let val () = check_con_kind ctx c1 KndType 
                            "synth_con_kind: TyArrow with non-type component"
                    val () = check_con_kind ctx c2 KndType
                            "synth_con_kind: TyArrow with non-type component"
                in
                    KndType
                end
          | TyRef c =>
                let val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyRef with non-type argument"
                in
                    KndType
                end
          | TyTagged => KndType
          | TyTag c =>
                let val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyTag with non-type argument"
                in
                    KndType
                end
          | TySum cs =>
                let val () = List.app (fn c => check_con_kind ctx c KndType
                                               ("synth_con_kind: TySum "
                                               ^ "with non-type component"))
                                      cs
                in
                    KndType
                end
          | TyProd cs =>
                let val () = List.app (fn c => check_con_kind ctx c KndType
                                               ("synth_con_kind: TyProd "
                                               ^ "with non-type component"))
                                      cs
                in
                    KndType
                end
          | TyForall (cx, k, c) =>
                let val (cx, c) = alpha_vary (cx, c) replace_con_con
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val () = check_con_kind ctx c KndType
                             "synth_con_kind: TyForall with non-type body"
                in
                    KndType
                end

    and check_con_kind ctx con knd s =
        let val knd' = synth_con_kind ctx con
        in
            if equiv_kind knd' knd then
                ()
            else
                raise Check s
        end

    (* expansion for recursive kinds, a la Dreyer, but only for kinds T^n *)
    and expand ctx recpath =
        let (* decomposes a path into its elimination context and its root
               if [ decompose_path path ] = (elim_ctx, root),
               then [ elim_ctx root ] = path *)
            fun decompose_path path =
                case path of
                    ConProj1 cp =>
                        let val (elim_ctx, root) = decompose_path cp
                        in
                            (fn c => ConProj1 (elim_ctx c), root)
                        end
                  | ConProj2 cp =>
                        let val (elim_ctx, root) = decompose_path cp
                        in
                            (fn c => ConProj2 (elim_ctx c), root)
                        end
                  | _ => (fn c => c, path)
            val () = check_con_kind ctx recpath KndType
                     "expand: recursive path is not a type"
            val (elim_ctx, c_mu) = decompose_path recpath
        in
            case c_mu of
                ConMu (cx, k, c) => elim_ctx (subst_con_con c_mu cx c)
              | _ => raise Check "expand: recursive path not rooted at ConMu"
        end

    fun synth_exp_con ctx exp =
        case exp of
            ExVar x => 
                (case Ctx.lookup_expvar ctx x of
                    SOME ty => ty
                  | NONE => raise Check ("synth_exp_con: unbound expvar, "
                                        ^ Variable.toString x))
          | ExLit l => TyLit (Literal.lit_ty l)
          | ExUnOp(oper, e) =>
                let val (cin, cout) = Literal.unop_tys oper TyLit tyBool tyUnit
                    val () = check_exp_con ctx e cin
                             "synth_exp_con: ExUnOp argument incorrect"
                in
                  cout
                end
          | ExBinOp(e1, oper, e2) =>
                let val (cin1, cin2, cout) =
                          Literal.binop_tys oper TyLit tyBool tyUnit
                    val () = check_exp_con ctx e1 cin1
                             "synth_exp_con: ExBinOp first argument incorrect"
                    val () = check_exp_con ctx e2 cin2
                             "synth_exp_con: ExBinOp second argument incorrect"
                in
                    cout
                end
          | ExFix (i, fbnds) =>
                let (* every function body is checked in an enriched scope
                       including bindings for all the functions *)
                       (* XXX should technically be doing some massive
                          alpha-varying, but this should work since no
                          classifier can depend on expression variables *)
                    val ctx0 = ctx (* save original context *)
                    val ctx = List.foldl
                                (fn ((f, _, c, c', _), ctx) =>
                                      Ctx.extend ctx
                                      (Ctx.DecExpCon (f, TyArrow (c, c'))))
                                ctx
                                fbnds
                    (* each individual function body is checked in a scope
                       enriched with the type of its formal parameter *)
                    fun check_fbnd ctx (f, x, c, cbody, ebody) =
                        let val ctx = Ctx.extend ctx (Ctx.DecExpCon (x, c))
                            val () = check_con_kind ctx0 c KndType
                                     "synth_exp_con: ExFix domain not a type"
                            val () = check_con_kind ctx0 cbody KndType
                                     "synth_exp_con: ExFix co-domain not a type"
                            val () = check_exp_con ctx ebody cbody
                                   "synth_exp_con: function body has wrong type"
                        in
                            ()
                        end
                    val () = List.app (check_fbnd ctx) fbnds
                in
                    (* the resulting type is the type of the i-th function *)
                    case nth (fbnds, i-1) of
                        SOME (_, _, c, c', _) => TyArrow (c, c')
                      | NONE => raise Check ("synth_exp_con: "
                                            ^ "ExFix subscript out of bounds")
                end
          | ExApp (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf c1 of
                        TyArrow (c1dom, c1cod) =>
                            (* XXX use check_exp_con instead? *)
                            let val c2 = synth_exp_con ctx e2
                                val () = check_equiv_con ctx c1dom c2 KndType
                                         ("synth_exp_con: function argument "
                                         ^ "doesn't match function domain")
                            in
                                c1cod
                            end
                      | _ => raise Check
                             "synth_exp_con: application of a non-function"
                end
          | ExHandle (e1, (x, e2)) =>
                let val c1 = synth_exp_con ctx e1
                    val (x, e2) = alpha_vary (x, e2) replace_exp_exp
                    val ctx' = Ctx.extend ctx (Ctx.DecExpCon (x, TyTagged))
                    val c2 = synth_exp_con ctx' e2
                    val () = check_equiv_con ctx c1 c2 KndType
                             "synth_exp_con: ExHandle handler body has wrong return type"
                in
                    c1
                end
          | ExRaise (c, e) =>
                let val () = check_exp_con ctx e TyTagged
                             "synth_exp_con: raise arg not of type TyTagged"
                    val () = check_con_kind ctx c KndType
                             "synth_exp_con: con arg to raise not a type"
                in
                    c
                end
          | ExRef e => TyRef(synth_exp_con ctx e)
          | ExGet e =>
                let val c = synth_exp_con ctx e
                in
                    case whnf c of
                        TyRef c' => c'
                      | _ => raise Check
                             "synth_exp_con: ExGet applied to a non-ref"
                end
          | ExSet (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf c1 of
                        TyRef c2 =>
                            let val () = check_exp_con ctx e2 c2
                                         ("synth_exp_con: ExSet arg has "
                                         ^ "wrong type")
                            in
                                tyUnit
                            end
                      | _ => raise Check
                             "synth_exp_con: ExSet applied to a non-ref"
                end
          (* see Dreyer's dissertation for the Roll/Unroll rules *)
          | ExRoll (c, e) =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExRoll at non-type"
                    val recpath = whnf c
                    val c_expanded = expand ctx recpath
                    val () = check_exp_con ctx e c_expanded
                             "synth_exp_con: ExRoll arg has wrong type"
                in
                    recpath
                end
          | ExUnroll e =>
                let val c = synth_exp_con ctx e
                    val recpath = whnf c
                    val c_expanded = expand ctx recpath
                in
                    c_expanded
                end
          | ExProd es => TyProd (map (synth_exp_con ctx) es)
          | ExProj (i, e) =>
                let val c = synth_exp_con ctx e
                in
                    case whnf c of
                        TyProd tys =>
                            (case nth (tys, i-1) of
                                NONE => raise Check ("synth_exp_con: "
                                        ^ "ExProj subscript out of bounds, "
                                        ^ Int.toString i)
                              | SOME ty => ty)
                      | _ => raise Check
                             "synth_exp_con: projection from a non-product"
                end
          | ExSumInj (c, i, e) =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExSumInj at non-type"
                    val c = whnf c
                in
                    case c of
                        TySum tys =>
                            (case nth (tys, i-1) of
                                SOME ty =>
                                    let val () = check_exp_con ctx e ty
                                                 ("synth_exp_con: ExSumInj "
                                                 ^ "arg has wrong type")
                                    in
                                        c
                                    end
                              | NONE => raise Check ("synth_exp_con: "
                                        ^ "ExSumInj subscript out of bounds, "
                                        ^ Int.toString i))
                      | _ => raise Check
                             "synth_exp_con: ExSumInj into non-TySum type"
                end
          | ExCase (c_cod, e, es) =>
                let val c_sum = synth_exp_con ctx e
                in
                    case whnf c_sum of
                        TySum tys =>
                            let val () =
                                ListPair.app
                                    (fn (c_dom, e) =>
                                      check_exp_con ctx e
                                       (TyArrow (c_dom, c_cod))
                                       "synth_exp_con: type mismatch in ExCase")
                                    (tys, es)
                            in
                                c_cod
                            end
                      | _ => raise Check
                             "synth_exp_con: ExCase over non-TySum type"
                end
          | ExTagInj (e1, e2) =>
                let val c1 = synth_exp_con ctx e1
                in
                    case whnf c1 of
                        TyTag c2 =>
                            let val () = check_exp_con ctx e2 c2
                                         ("synth_exp_con: ExTagInj data has "
                                         ^ "incorrect type")
                            in
                                TyTagged
                            end
                      | _ => raise Check "synth_exp_con: ExTagInj into non-tag"
                end
          | ExNewTag c =>
                let val () = check_con_kind ctx c KndType
                             "synth_exp_con: ExNewTag carrying non-type"
                in
                    TyTag c
                end
          | ExIfTagOf (e1, e2, e3, e4) =>
                let val () = check_exp_con ctx e1 TyTagged
                             "synth_exp_con: ExIfTagOf on non-TyTagged"
                    val c_data =
                        case whnf (synth_exp_con ctx e2) of
                            TyTag c => c
                          | _ => raise Check ("synth_exp_con: ExIfTagOf "
                                             ^ "comparison exp not a TyTag")
                    val c_res =
                        case whnf (synth_exp_con ctx e3) of
                            TyArrow (c_data', c) =>
                                let val () = check_equiv_con ctx c_data c_data'
                                             KndType
                                             ("synth_exp_con: ExIfTagOf `then' "
                                             ^ "function has incorrect domain")
                                in
                                    c
                                end
                          | _ => raise Check ("synth_exp_con: ExIfTagOf `then' "
                                             ^ "exp not a function")
                    val () = check_exp_con ctx e4 c_res
                             ("synth_exp_con: ExIfTagOf `else' exp "
                             ^ "has incorrect type")
                in
                    c_res
                end
          | ExConAbs (cx, k, e) =>
                let val (cx, e) = alpha_vary (cx, e) replace_con_exp
                    val ctx = Ctx.extend ctx (Ctx.DecConKind (cx, k))
                    val c = synth_exp_con ctx e
                in
                    TyForall (cx, k, c)
                end
          | ExConApp (e, c_arg) =>
                let val c_fun = synth_exp_con ctx e
                in
                    case whnf c_fun of
                        TyForall (cx, k, c) =>
                            let val () = check_con_kind ctx c_arg k
                                         ("synth_con_kind: ExConApp arg "
                                         ^ "has wrong kind")
                            in
                                subst_con_con c_arg cx c
                            end
                      | _ => raise Check "con-application of a non-forall"
                end
          | ExLet (ex, e, e') =>
                let val c = synth_exp_con ctx e
                    val (ex, e') = alpha_vary (ex, e') replace_exp_exp
                    val ctx = Ctx.extend ctx (Ctx.DecExpCon (ex, c))
                    val c' = synth_exp_con ctx e'
                in
                    c'
                end

    and check_exp_con ctx exp con s =
        let val con' = synth_exp_con ctx exp
        in
            equiv_con ctx con con' KndType
            handle Equiv => raise Check s
        end
  end
